from __future__ import annotations

import enum
import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, ClassVar, Generic, TypeVar, cast

_COLON = re.compile(r":")


def _no_colon(value: str, *, field_name: str) -> str:
    if _COLON.search(value):
        raise ValueError(f"{field_name} may not contain ':' - got {value!r}")
    return value


class Endpoint(enum.Enum):
    AUTHORIZE = "/api/v1/authorize"


class Decision(str, enum.Enum):
    ALLOW = "Allow"
    DENY = "Deny"


@dataclass(slots=True, frozen=True)
class QualifiedId:
    id: str
    namespace: list[str] = field(default_factory=list)

    def __post_init__(self):
        _no_colon(self.id, field_name="QualifiedId.id")
        for ns in self.namespace:
            _no_colon(ns, field_name="QualifiedId.namespace element")

    def to_api(self) -> dict[str, Any]:
        return {"id": self.id, "namespace": self.namespace}

    def __str__(self) -> str:
        if self.namespace:
            return f"{'::'.join(self.namespace)}::{self.id}"
        return self.id


@dataclass(slots=True, frozen=True)
class Group:
    id: QualifiedId

    def to_api(self) -> dict[str, Any]:
        return self.id.to_api()

    @classmethod
    def new(
        cls,
        id: str,
        namespace: list[str] | None = None,
    ) -> Group:
        """Create a new Group with a QualifiedId."""
        return Group(id=QualifiedId(id=id, namespace=namespace or []))

    def __str__(self) -> str:
        return str(self.id)


@dataclass(slots=True, frozen=True)
class Action:
    id: QualifiedId

    def to_api(self) -> dict[str, Any]:
        return self.id.to_api()

    @classmethod
    def new(
        cls,
        id: str,
        namespace: list[str] | None = None,
    ) -> Action:
        """Create a new Action."""
        return Action(id=QualifiedId(id=id, namespace=namespace or []))

    def __str__(self) -> str:
        return str(self.id)


@dataclass(slots=True, frozen=True)
class User:
    id: QualifiedId
    groups: list[Group] = field(default_factory=list)

    def to_api(self) -> dict[str, Any]:
        return {
            **self.id.to_api(),
            "groups": [g.to_api() for g in self.groups],
        }

    @classmethod
    def new(
        cls,
        id: str,
        namespace: list[str] | None = None,
        groups: list[str] | None = None,
    ) -> User:
        """Create a new User.

        Note that this interface assumes that groups share the same namespace as the user.
        If groups are in a different namespace, you should create the object manually.

        Args:
            id: The user ID.
            namespace: Optional namespace for the user ID.
            groups: Optional list of group IDs. If provided, they will be converted to Group objects
                     with the same namespace as the user.
        Returns:
            A User object with the given ID, namespace, and groups.
        """
        return User(
            id=QualifiedId(id=id, namespace=namespace or []),
            groups=[Group.new(id=g, namespace=namespace or []) for g in groups or []],
        )

    def __str__(self) -> str:
        return str(self.id)


Principal = User | Group


class ResourceAttributeType(str, enum.Enum):
    STRING = "String"
    NUMBER = "Long"
    BOOLEAN = "Boolean"
    IP = "Ip"


@dataclass(slots=True, frozen=True)
class ResourceAttribute:
    type: ResourceAttributeType
    value: str

    def to_api(self) -> dict[str, Any]:
        if self.type == ResourceAttributeType.BOOLEAN:
            # Convert "true"/"false" strings to actual booleans for the API
            val = self.value.lower() == "true"
            return {"type": self.type.value, "value": val}
        elif self.type == ResourceAttributeType.NUMBER:
            # Convert numeric strings to actual floats for the API
            val = float(self.value)
            return {"type": self.type.value, "value": val}

        return {"type": self.type.value, "value": self.value}

    @classmethod
    def new(
        cls, value: str, type: ResourceAttributeType | None = None
    ) -> ResourceAttribute:
        """Create a new ResourceAttribute."""
        if type is None:
            type = ResourceAttributeType.STRING

        return cls(type=type, value=value)


@dataclass(slots=True, frozen=True)
class Resource:
    kind: str
    id: str
    attrs: dict[str, ResourceAttribute]

    def __post_init__(self):
        _no_colon(self.kind, field_name="Resource.kind")
        if not self.attrs:
            raise ValueError("Resource.attrs cannot be empty")

    def to_api(self) -> dict[str, Any]:
        return {
            "kind": self.kind,
            "id": self.id,
            "attrs": {k: v.to_api() for k, v in self.attrs.items()},
        }

    @classmethod
    def new(cls, kind: str, id: str, attrs: dict[str, ResourceAttribute]) -> Resource:
        """Create a new Resource with kind and attributes."""
        return cls(kind=kind, id=id, attrs=attrs)


@dataclass(slots=True, frozen=True)
class Request:
    principal: Principal
    action: Action
    resource: Resource
    id: str | None = None

    def to_api(self) -> dict[str, Any]:
        # Principal: User already returns {"User": {...}}.
        # Group should be wrapped as {"Group": {...}} here.
        if isinstance(self.principal, User):
            principal_payload: dict[str, Any] = {"User": self.principal.to_api()}
        elif isinstance(self.principal, Group):  # type: ignore[unreachable]
            principal_payload = {"Group": self.principal.to_api()}
        else:
            raise TypeError(
                f"Unsupported principal type: {type(self.principal).__name__}"
            )

        result: dict[str, Any] = {
            "principal": principal_payload,
            "action": self.action.to_api(),
            "resource": self.resource.to_api(),
        }
        if self.id is not None:
            result["id"] = self.id
        return result


def as_api(obj: Request | dict[str, Any]) -> dict[str, Any]:
    return obj.to_api() if isinstance(obj, Request) else obj


@dataclass(slots=True, frozen=True)
class AuthorizedResponseBrief:
    decision: Decision

    _KEYS: ClassVar[tuple[str, ...]] = ("decision",)

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizedResponseBrief:
        # will KeyError if missing, or propagate other types
        dec = data.get("decision", data.get("desicion"))
        if dec is None:
            raise KeyError("decision")
        return cls(decision=Decision(dec))

    def is_allowed(self) -> bool:
        return self.decision == Decision.ALLOW

    def is_denied(self) -> bool:
        return self.decision == Decision.DENY


@dataclass(slots=True, frozen=True)
class PermitPolicy:
    literal: str
    json: dict[str, Any]

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> PermitPolicy:
        return cls(literal=data["literal"], json=data["json"])


@dataclass(slots=True, frozen=True)
class PolicyVersion:
    hash: str
    loaded_at: datetime

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> PolicyVersion:
        return cls(
            hash=data["hash"],
            loaded_at=datetime.fromisoformat(data["loaded_at"].replace("Z", "+00:00")),
        )


@dataclass(slots=True, frozen=True)
class AuthorizedResponseDetailed:
    # Either decision == Decision.DENY or Decision.ALLOW with PermitPolicy
    decision: Decision
    policy: PermitPolicy | None
    version: PolicyVersion | None = None

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizedResponseDetailed:
        dec = data.get("decision", data.get("desicion")) # Temporary typo support
        if dec is None:
            raise KeyError("decision")

        # 1) Is it a simple Deny (old format)?
        if dec == Decision.DENY.value:
            version_blob = data.get("version")
            version = PolicyVersion.from_api(version_blob) if version_blob else None
            return cls(decision=Decision.DENY, policy=None, version=version)

        # 1b) Is it a simple Allow with top-level policy/version?
        if dec == Decision.ALLOW.value:
            if "policy" not in data:
                raise ValueError(f"Allow decision missing policy: {data!r}")
            policy_blob = data["policy"]
            version_blob = data.get("version")
            version = PolicyVersion.from_api(version_blob) if version_blob else None
            return cls(
                decision=Decision.ALLOW,
                policy=PermitPolicy.from_api(policy_blob),
                version=version,
            )

        # 2) Is it a Deny with version (new format)?
        if isinstance(dec, dict) and Decision.DENY.value in dec:
            deny_dict = cast(dict[str, Any], dec[Decision.DENY.value])
            version_blob = deny_dict.get("version")
            version = PolicyVersion.from_api(version_blob) if version_blob else None
            return cls(
                decision=Decision.DENY,
                policy=None,
                version=version,
            )

        # 3) If it's a dict with an "Allow" key, pull the policy and optional version
        if isinstance(dec, dict) and Decision.ALLOW.value in dec:
            if not isinstance(dec[Decision.ALLOW.value], dict):
                raise ValueError(f"Malformed Allow decision: {dec!r}")

            allow_dict = cast(dict[str, Any], dec[Decision.ALLOW.value])

            if "policy" not in allow_dict:
                raise ValueError(f"Allow decision missing policy: {dec!r}")

            policy_blob = allow_dict["policy"]
            version_blob = allow_dict.get("version")
            version = PolicyVersion.from_api(version_blob) if version_blob else None
            return cls(
                decision=Decision.ALLOW,
                policy=PermitPolicy.from_api(policy_blob),
                version=version,
            )

        # 4) Otherwise it's malformed
        raise ValueError(f"Unrecognized decision shape: {dec!r}")

    def is_allowed(self) -> bool:
        return self.decision == Decision.ALLOW

    def is_denied(self) -> bool:
        return self.decision == Decision.DENY

    def policy_literal(self) -> str | None:
        """Return the policy literal if available, otherwise None."""
        return self.policy.literal if self.policy else None

    def policy_json(self) -> dict[str, Any] | None:
        """Return the policy JSON if available, otherwise None."""
        return self.policy.json if self.policy else None

    def version_hash(self) -> str | None:
        """Return the policy version hash if available, otherwise None."""
        return self.version.hash if self.version else None

    def version_loaded_at(self) -> datetime | None:
        """Return the policy version loaded_at timestamp if available, otherwise None."""
        return self.version.loaded_at if self.version else None


# Generic type variables for authorization results and responses
ResultT = TypeVar('ResultT', bound='AuthorizedResponseBrief | AuthorizedResponseDetailed')
T = TypeVar('T', bound='AuthorizeResultBrief | AuthorizeResultDetailed')


@dataclass(slots=True, frozen=True)
class AuthorizeResultBase(Generic[ResultT]):
    """Base class for a single result from the authorize endpoint."""

    index: int
    id: str | None
    status: str
    result: ResultT | None = None
    error: str | None = None

    def is_success(self) -> bool:
        """Check if this result is successful."""
        return self.status == "success"

    def is_failed(self) -> bool:
        """Check if this result failed."""
        return self.status == "failed"

    def get_decision(self) -> Decision | None:
        """Get the decision if successful, otherwise None."""
        return self.result.decision if self.result else None

    def is_allowed(self) -> bool:
        """Check if the decision is Allow."""
        return self.result.is_allowed() if self.result else False

    def is_denied(self) -> bool:
        """Check if the decision is Deny."""
        return self.result.is_denied() if self.result else False


@dataclass(slots=True, frozen=True)
class AuthorizeResultBrief(AuthorizeResultBase[AuthorizedResponseBrief]):
    """A single result from the authorize endpoint (brief detail level)."""

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizeResultBrief:
        index = data["index"]
        result_id = data.get("id")
        status = data["status"]
        error = data.get("error")

        result = None
        if status == "success" and "result" in data:
            result = AuthorizedResponseBrief.from_api(data["result"])

        return cls(index=index, id=result_id, status=status, result=result, error=error)


@dataclass(slots=True, frozen=True)
class AuthorizeResultDetailed(AuthorizeResultBase[AuthorizedResponseDetailed]):
    """A single result from the authorize endpoint (detailed level)."""

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizeResultDetailed:
        index = data["index"]
        result_id = data.get("id")
        status = data["status"]
        error = data.get("error")

        result = None
        if status == "success" and "result" in data:
            result = AuthorizedResponseDetailed.from_api(data["result"])

        return cls(index=index, id=result_id, status=status, result=result, error=error)

    def policy_literal(self) -> str | None:
        """Return the policy literal if available, otherwise None."""
        return self.result.policy_literal() if self.result else None

    def policy_json(self) -> dict[str, Any] | None:
        """Return the policy JSON if available, otherwise None."""
        return self.result.policy_json() if self.result else None

    def version_hash(self) -> str | None:
        """Return the policy version hash if available, otherwise None."""
        return self.result.version_hash() if self.result else None

    def version_loaded_at(self) -> datetime | None:
        """Return the policy version loaded_at timestamp if available, otherwise None."""
        return self.result.version_loaded_at() if self.result else None


# Generic type variables for authorization results and responses
ResultT = TypeVar('ResultT', bound='AuthorizedResponseBrief | AuthorizedResponseDetailed')
T = TypeVar('T', bound='AuthorizeResultBrief | AuthorizeResultDetailed')


@dataclass(slots=True, frozen=True)
class AuthorizeResponseBase(Generic[T]):
    """Base class for batch responses from the authorize endpoint."""

    results: list[T]
    version: PolicyVersion
    successful: int
    failed: int

    def __iter__(self):
        """Iterate over results."""
        return iter(self.results)

    def __len__(self):
        """Return the number of results."""
        return len(self.results)

    def __getitem__(self, index: int) -> T:
        """Get result by index."""
        return self.results[index]

    def get_by_id(self, request_id: str) -> T | None:
        """Get result by client-provided request ID."""
        for result in self.results:
            if result.id == request_id:
                return result
        return None

    def denied_count(self) -> int:
        """Return the number of denied results."""
        return sum(1 for r in self.results if r.is_success() and r.is_denied())

    def allowed_count(self) -> int:
        """Return the number of allowed results."""
        return sum(1 for r in self.results if r.is_success() and r.is_allowed())

    def all_allowed(self) -> bool:
        """Check if all successful results are allowed."""
        return all(r.is_allowed() for r in self.results if r.is_success())


@dataclass(slots=True, frozen=True)
class AuthorizeResponseBrief(AuthorizeResponseBase[AuthorizeResultBrief]):
    """Batch response from the authorize endpoint (brief detail level)."""

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizeResponseBrief:
        results = [AuthorizeResultBrief.from_api(r) for r in data.get("results", [])]
        version = PolicyVersion.from_api(data["version"])
        successful = data.get("successful", 0)
        failed = data.get("failed", 0)

        return cls(
            results=results, version=version, successful=successful, failed=failed
        )


@dataclass(slots=True, frozen=True)
class AuthorizeResponseDetailed(AuthorizeResponseBase[AuthorizeResultDetailed]):
    """Batch response from the authorize endpoint (detailed level)."""

    @classmethod
    def from_api(cls, data: dict[str, Any]) -> AuthorizeResponseDetailed:
        results = [AuthorizeResultDetailed.from_api(r) for r in data.get("results", [])]
        version = PolicyVersion.from_api(data["version"])
        successful = data.get("successful", 0)
        failed = data.get("failed", 0)

        return cls(
            results=results, version=version, successful=successful, failed=failed
        )



import subprocess
import time

import httpx
import pytest

from treetop_client.client import TreeTopClient
from treetop_client.models import (Action, Decision, Request, Resource,
                                   ResourceAttribute, ResourceAttributeType,
                                   User)

pytestmark = pytest.mark.integration

PORT = 10101
NAMESPACE = ["DNS"]


def make_host_resource(
    host_id: str, ip: str = "10.0.0.1"
) -> Resource:
    """Create a Host resource with name and IP attributes."""
    return Resource.new(
        kind="Host",
        id=host_id,
        attrs={
            "name": ResourceAttribute.new(host_id),
            "ip": ResourceAttribute.new(ip, type=ResourceAttributeType.IP),
        },
    )


def make_request(
    principal: str,
    action: str,
    host_id: str,
    groups: list[str] | None = None,
    request_id: str | None = None,
    ip: str = "10.0.0.1",
) -> Request:
    """Create an authorization request for a Host resource."""
    return Request(
        principal=User.new(principal, NAMESPACE, groups),
        action=Action.new(action, NAMESPACE),
        resource=make_host_resource(host_id, ip),
        id=request_id,
    )


@pytest.fixture
def client(docker_compose_up_down) -> TreeTopClient: # type: ignore[reportMissingParameterType]
    """Create a TreeTopClient connected to the test server."""
    return TreeTopClient(base_url=f"http://localhost:{PORT}")


@pytest.fixture(scope="session")
def docker_compose_up_down(tmp_path_factory: pytest.TempPathFactory):
    """
    Spin up the server via docker-compose.integration.yml before tests,
    tear it down afterwards.
    """
    # bring up
    subprocess.check_call(
        ["docker", "compose", "-f", "docker-compose.integration.yml", "up", "-d"]
    )
    # wait for the server to be ready
    for _ in range(10):
        try:
            resp = httpx.get(f"http://localhost:{PORT}/api/v1/policies", timeout=1.0)
            if resp.status_code == 200:
                entries = resp.json().get("policies", {}).get("entries", 0)
                if entries:
                    break
                else:
                    time.sleep(1)
        except Exception:
            time.sleep(1)
    else:
        pytest.skip("policy-server did not start in time")
    yield
    # tear down
    subprocess.call(
        ["docker", "compose", "-f", "docker-compose.integration.yml", "down"]
    )


@pytest.mark.parametrize(
    "principal, groups, action, expected",
    [
        (
            "alice",
            ["admins", "users"],
            "create_host",
            True,
        ),
        (
            "bob",
            ["users"],
            "create_host",
            False,
        ),
        (
            "alice",
            ["admins", "users"],
            "view_host",
            True,
        ),
        (
            "bob",
            ["users"],
            "view_host",
            True,
        ),
        (
            "alice",
            ["admins", "users"],
            "delete_host",
            True,
        ),
        (
            "bob",
            ["users"],
            "delete_host",
            False,
        ),
    ],
)
def test_live_check_allows_user(
    principal: str,
    groups: list[str],
    action: str,
    expected: bool,
    client: TreeTopClient,
):
    req = make_request(principal, action, "host.example.com", groups)
    resp = client.check(req)
    if expected:
        assert resp.is_allowed()
        assert resp.decision == Decision.ALLOW
    else:
        assert resp.is_denied()
        assert resp.decision == Decision.DENY


@pytest.mark.parametrize(
    "resource_kind, id, attrs",
    [
        ("Generic", "12345", {"kind": ResourceAttribute.new("Any")}),
        (
            "Host",
            "host.example.com",
            {
                "name": ResourceAttribute.new("host.example.com"),
                "ip": ResourceAttribute.new("10.0.0.1", type=ResourceAttributeType.IP),
            },
        ),
    ],
)
def test_live_check_allows_super_bare(
    resource_kind: str,
    id: str,
    attrs: dict[str, ResourceAttribute],
    client: TreeTopClient,
):
    req = Request(
        principal=User.new("super"),
        action=Action.new("any"),
        resource=Resource.new(resource_kind, id, attrs),
    )
    resp = client.check(req)
    assert resp.is_allowed()
    assert resp.decision == Decision.ALLOW


def test_live_check_allow_detailed(
    client: TreeTopClient,
):
    req = make_request("alice", "view_host", "host.example.com", ["admins"])
    resp = client.check_detailed(req)
    assert resp.is_allowed()
    assert resp.decision == Decision.ALLOW
    assert resp.policy is not None
    assert (
        resp.policy_literal()
        == """@id("DNS.admins_policy")
permit (
    principal in DNS::Group::"admins",
    action in
        [DNS::Action::"create_host",
         DNS::Action::"delete_host",
         DNS::Action::"view_host",
         DNS::Action::"edit_host"],
    resource is Host
);"""
    )
    # Verify version information is present
    assert resp.version_hash() is not None
    assert resp.version_loaded_at() is not None


# Batch authorization tests
def test_live_batch_authorize_multiple_requests(
    client: TreeTopClient,
):
    """Test batch authorize with multiple requests."""
    requests = [
        make_request("alice", "create_host", "host1.example.com", ["admins"], "req-alice-create", "10.0.0.1"),
        make_request("bob", "create_host", "host2.example.com", ["users"], "req-bob-create", "10.0.0.2"),
        make_request("bob", "view_host", "host3.example.com", ["users"], "req-bob-view", "10.0.0.3"),
    ]

    response = client.authorize(requests)

    # Verify counts
    assert response.successful == 3
    assert response.failed == 0
    assert len(response) == 3

    # Verify results by index
    assert response[0].is_success()
    assert response[0].is_allowed()  # alice can create_host (admin)

    assert response[1].is_success()
    assert response[1].is_denied()  # bob cannot create_host (not admin)

    assert response[2].is_success()
    assert response[2].is_allowed()  # bob can view_host (user)


def test_live_batch_authorize_lookup_by_id(
    client: TreeTopClient,
):
    """Test batch authorize with lookup by request ID."""
    requests = [
        make_request("alice", "delete_host", "dangerous.example.com", ["admins"], "admin-delete", "10.0.0.99"),
        make_request("bob", "delete_host", "restricted.example.com", ["users"], "user-delete", "10.0.0.88"),
    ]

    response = client.authorize(requests)

    # Lookup by ID
    admin_result = response.get_by_id("admin-delete")
    assert admin_result is not None
    assert admin_result.is_allowed()  # alice can delete_host (admin)

    user_result = response.get_by_id("user-delete")
    assert user_result is not None
    assert user_result.is_denied()  # bob cannot delete_host (user)

    # Verify non-existent ID returns None
    nonexistent = response.get_by_id("nonexistent-id")
    assert nonexistent is None


def test_live_batch_authorize_detailed_multiple_requests(
    client: TreeTopClient,
):
    """Test batch authorize_detailed with multiple requests."""
    requests = [
        make_request("alice", "create_host", "host-a.example.com", ["admins"], "alice-create-detailed", "10.0.0.1"),
        make_request("bob", "view_host", "host-b.example.com", ["users"], "bob-view-detailed", "10.0.0.2"),
    ]

    response = client.authorize_detailed(requests)

    # Verify batch metadata
    assert response.successful == 2
    assert response.failed == 0

    # Verify detailed results by index
    alice_result = response[0]
    assert alice_result.is_allowed()
    alice_policy = alice_result.policy_literal()
    assert alice_policy is not None
    assert "admins_policy" in alice_policy
    assert alice_result.version_hash() is not None

    bob_result = response[1]
    assert bob_result.is_allowed()
    bob_policy = bob_result.policy_literal()
    assert bob_policy is not None
    assert "users_policy" in bob_policy
    assert bob_result.version_hash() is not None


def test_live_batch_authorize_detailed_lookup_by_id(
    client: TreeTopClient,
):
    """Test batch authorize_detailed with lookup by ID."""
    requests = [
        Request(
            principal=User.new("super"),
            action=Action.new("any"),
            resource=make_host_resource("super-host.example.com", "10.0.0.100"),
            id="super-admin-any",
        ),
        make_request("bob", "create_host", "bob-host.example.com", ["users"], "bob-create-detailed", "10.0.0.50"),
    ]

    response = client.authorize_detailed(requests)

    # Lookup by ID
    super_result = response.get_by_id("super-admin-any")
    assert super_result is not None
    assert super_result.is_allowed()  # super can do anything
    assert super_result.policy_literal() is not None
    assert super_result.version_hash() is not None
    assert super_result.version_loaded_at() is not None

    bob_result = response.get_by_id("bob-create-detailed")
    assert bob_result is not None
    assert bob_result.is_denied()  # bob cannot create_host (user)
    assert bob_result.policy_literal() is None  # Deny has no policy


def test_live_batch_authorize_iteration(
    client: TreeTopClient,
):
    """Test iterating over batch results and filtering."""
    requests = [
        make_request("alice", "create_host", "host-iter-1.example.com", ["admins"], "iter-1", "10.1.0.1"),
        make_request("bob", "create_host", "host-iter-2.example.com", ["users"], "iter-2", "10.1.0.2"),
        make_request("alice", "view_host", "host-iter-3.example.com", ["admins"], "iter-3", "10.1.0.3"),
    ]

    response = client.authorize(requests)

    # Test iteration
    decisions = [result.get_decision() for result in response]
    assert len(decisions) == 3
    assert decisions[0] == Decision.ALLOW  # alice create_host
    assert decisions[1] == Decision.DENY   # bob create_host
    assert decisions[2] == Decision.ALLOW  # alice view_host

    # Test filtering authorized requests
    allowed_count = sum(1 for result in response if result.is_allowed())
    assert allowed_count == 2

    # Test filtering denied requests
    denied_count = sum(1 for result in response if result.is_denied())
    assert denied_count == 1


def test_live_batch_large_query_ordering_and_performance(
    client: TreeTopClient,
) -> None:
    """Test batch authorize with 50+ requests - verify ordering and performance.
    
    Performance Notes:
    - Tests the server's ability to handle large batches efficiently
    - Verifies that results maintain the same order as requests (critical!)
    - Measures end-to-end response time for bulk operations
    - Useful for benchmarking and understanding scalability limits
    """
    import time as time_module

    # Create 60 requests with various combinations
    # Alternating between admins/users, and different actions
    num_requests = 60
    requests: list[tuple[Request, bool]] = []
    
    for i in range(num_requests):
        # Alternate: admins get all operations, users get view/edit only
        if i % 3 == 0:
            principal = "alice"
            groups = ["admins"]
            action = "create_host"
            expected_allowed = True
        elif i % 3 == 1:
            principal = "bob"
            groups = ["users"]
            action = "create_host"
            expected_allowed = False  # users can't create
        else:
            principal = "bob"
            groups = ["users"]
            action = "view_host"
            expected_allowed = True  # users can view
        
        req = make_request(
            principal=principal,
            action=action,
            host_id=f"host-stress-{i}.example.com",
            groups=groups,
            request_id=f"stress-{i}",
            ip=f"10.2.{i // 256}.{i % 256}",  # Generate diverse IPs
        )
        requests.append((req, expected_allowed))
    
    # Time the batch operation
    start_time = time_module.time()
    response = client.authorize([req for req, _ in requests])
    elapsed_time = time_module.time() - start_time
    
    # Verify response counts
    assert len(response) == num_requests
    assert response.successful == num_requests
    assert response.failed == 0
    
    # Verify all results came back in the correct order
    for i, (_, expected_allowed) in enumerate(requests):
        result = response[i]
        assert result.index == i, f"Result {i} has wrong index: {result.index}"
        assert result.id == f"stress-{i}", f"Result {i} has wrong ID: {result.id}"
        assert result.is_success(), f"Result {i} should be successful"
        
        if expected_allowed:
            assert result.is_allowed(), f"Result {i} should be allowed (alice or bob viewing)"
        else:
            assert result.is_denied(), f"Result {i} should be denied (bob creating)"
    
    # Verify lookup by ID works for random samples
    stress_0_result = response.get_by_id("stress-0")
    assert stress_0_result is not None
    assert stress_0_result.is_allowed() is True
    
    stress_1_result = response.get_by_id("stress-1")
    assert stress_1_result is not None
    assert stress_1_result.is_denied() is True
    
    stress_59_result = response.get_by_id("stress-59")
    assert stress_59_result is not None
    assert stress_59_result.is_allowed() is True
    
    # Performance reporting
    requests_per_second = num_requests / elapsed_time
    print(f"\n==== Batch Authorization Performance ====")
    print(f"Total requests  : {num_requests}")
    print(f"Total time      : {elapsed_time:.3f} seconds")
    print(f"Throughput      : {requests_per_second:.1f} requests/second")
    print(f"Per-request time: {(elapsed_time / num_requests) * 1000:.1f} ms")
    print(f"=========================================\n")
    
    # Performance expectations (may vary by system)
    # Should handle 60 requests in < 0.5 seconds realistically
    assert elapsed_time < 0.5, f"Batch took too long: {elapsed_time:.2f}s for {num_requests} requests"



from __future__ import annotations

import contextlib
from collections.abc import Sequence
from typing import Any, Final

import httpx

from treetop_client.models import (AuthorizedResponseBrief,
                                   AuthorizedResponseDetailed,
                                   AuthorizeResponseBrief,
                                   AuthorizeResponseDetailed, Decision,
                                   Endpoint, Request, as_api)

_DEFAULT_LIMITS: Final = httpx.Limits(
    max_connections=100,
    max_keepalive_connections=20,
)


class _Singleton(type):
    _instance: TreeTopClient | None = None

    def __call__(cls, *args, **kwargs):  # type: ignore[override]
        """Ensure that only one instance of TreeTopClient exists."""
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class TreeTopClient(metaclass=_Singleton):
    def __init__(
        self,
        base_url: str = "http://localhost:9999",
        *,
        limits: httpx.Limits | None = None,
        timeout: float | httpx.Timeout = 5.0,
        verify: bool | str = True,
    ):
        self._sync_client = httpx.Client(
            base_url=base_url,
            limits=limits or _DEFAULT_LIMITS,
            timeout=timeout,
            verify=verify,
        )
        self._async_client = httpx.AsyncClient(
            base_url=base_url,
            limits=limits or _DEFAULT_LIMITS,
            timeout=timeout,
            verify=verify,
        )

    def _build_headers(self, correlation_id: str | None) -> dict[str, str] | None:
        """Build headers for the request, including a correlation ID if provided."""
        if not correlation_id:
            return None
        return {"X-Correlation-ID": correlation_id}

    def _sync_post(
        self,
        url: str,
        json_body: dict[str, Any],
        correlation_id: str | None = None,
        params: dict[str, str] | None = None,
    ) -> httpx.Response:
        """Synchronous POST request to the given URL with JSON body and optional correlation ID."""
        return self._sync_client.post(
            url,
            json=json_body,
            headers=self._build_headers(correlation_id),
            params=params,
        )

    async def _async_post(
        self,
        url: str,
        json_body: dict[str, Any],
        correlation_id: str | None = None,
        params: dict[str, str] | None = None,
    ) -> httpx.Response:
        """Asynchronous POST request to the given URL with JSON body and optional correlation ID."""
        return await self._async_client.post(
            url,
            json=json_body,
            headers=self._build_headers(correlation_id),
            params=params,
        )

    def authorize(
        self,
        requests: Request | dict[str, Any] | Sequence[Request | dict[str, Any]],
        correlation_id: str | None = None,
    ) -> AuthorizeResponseBrief:
        """Authorize one or more requests (brief detail level). Synchronous version.

        Args:
            requests: A single request or list of requests. Can be Request objects or dictionaries.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizeResponseBrief containing the batch results.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        request_list: list[dict[str, Any]] = []
        if isinstance(requests, (Request, dict)):
            request_list = [as_api(requests)]
        else:
            request_list = [as_api(req) for req in requests]
        resp = self._sync_post(
            Endpoint.AUTHORIZE.value,
            json_body={"requests": request_list},
            correlation_id=correlation_id,
        )
        resp.raise_for_status()
        return AuthorizeResponseBrief.from_api(resp.json())

    def authorize_detailed(
        self,
        requests: Request | dict[str, Any] | Sequence[Request | dict[str, Any]],
        correlation_id: str | None = None,
    ) -> AuthorizeResponseDetailed:
        """Authorize one or more requests (detailed with policy info). Synchronous version.

        Args:
            requests: A single request or list of requests. Can be Request objects or dictionaries.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizeResponseDetailed containing the batch results with policy info.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        request_list: list[dict[str, Any]] = []
        if isinstance(requests, (Request, dict)):
            request_list = [as_api(requests)]
        else:
            request_list = [as_api(req) for req in requests]
        resp = self._sync_post(
            Endpoint.AUTHORIZE.value,
            json_body={"requests": request_list},
            correlation_id=correlation_id,
            params={"detail": "full"},
        )
        resp.raise_for_status()
        return AuthorizeResponseDetailed.from_api(resp.json())

    async def aauthorize(
        self,
        requests: Request | dict[str, Any] | Sequence[Request | dict[str, Any]],
        correlation_id: str | None = None,
    ) -> AuthorizeResponseBrief:
        """Authorize one or more requests (brief detail level). Asynchronous version.

        Args:
            requests: A single request or list of requests. Can be Request objects or dictionaries.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizeResponseBrief containing the batch results.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        request_list: list[dict[str, Any]] = []
        if isinstance(requests, (Request, dict)):
            request_list = [as_api(requests)]
        else:
            request_list = [as_api(req) for req in requests]
        resp = await self._async_post(
            Endpoint.AUTHORIZE.value,
            json_body={"requests": request_list},
            correlation_id=correlation_id,
        )
        resp.raise_for_status()
        return AuthorizeResponseBrief.from_api(resp.json())

    async def aauthorize_detailed(
        self,
        requests: Request | dict[str, Any] | Sequence[Request | dict[str, Any]],
        correlation_id: str | None = None,
    ) -> AuthorizeResponseDetailed:
        """Authorize one or more requests (detailed with policy info). Asynchronous version.

        Args:
            requests: A single request or list of requests. Can be Request objects or dictionaries.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizeResponseDetailed containing the batch results with policy info.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        request_list: list[dict[str, Any]] = []
        if isinstance(requests, (Request, dict)):
            request_list = [as_api(requests)]
        else:
            request_list = [as_api(req) for req in requests]
        resp = await self._async_post(
            Endpoint.AUTHORIZE.value,
            json_body={"requests": request_list},
            correlation_id=correlation_id,
            params={"detail": "full"},
        )
        resp.raise_for_status()
        return AuthorizeResponseDetailed.from_api(resp.json())

    # Compatibility methods for single-request API (wraps batch API)
    def check(
        self, request: Request | dict[str, Any], correlation_id: str | None = None
    ) -> AuthorizedResponseBrief:
        """Check the given request. Synchronous version (compatibility wrapper).

        This method provides backward compatibility with the old single-request API.
        It wraps the new batch authorize endpoint.

        Args:
            request: The request to check, either as a Request object or a dictionary.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizedResponseBrief containing the result of the check.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        response = self.authorize(request, correlation_id=correlation_id)
        if not response.results:
            raise ValueError("No results returned from authorize endpoint")
        result = response.results[0]
        if result.status == "failed":
            raise RuntimeError(f"Authorization failed: {result.error}")
        return result.result or AuthorizedResponseBrief(Decision.DENY)

    def check_detailed(
        self, request: Request | dict[str, Any], correlation_id: str | None = None
    ) -> AuthorizedResponseDetailed:
        """Check the given request with detailed output. Synchronous version (compatibility wrapper).

        This method provides backward compatibility with the old single-request API.
        It wraps the new batch authorize_detailed endpoint.

        Args:
            request: The request to check, either as a Request object or a dictionary.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizedResponseDetailed containing the detailed result of the check.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        response = self.authorize_detailed(request, correlation_id=correlation_id)
        if not response.results:
            raise ValueError("No results returned from authorize endpoint")
        result = response.results[0]
        if result.status == "failed":
            raise RuntimeError(f"Authorization failed: {result.error}")
        return result.result or AuthorizedResponseDetailed(Decision.DENY, None, None)

    async def acheck(
        self, request: Request | dict[str, Any], correlation_id: str | None = None
    ) -> AuthorizedResponseBrief:
        """Check the given request. Asynchronous version (compatibility wrapper).

        This method provides backward compatibility with the old single-request API.
        It wraps the new batch aauthorize endpoint.

        Args:
            request: The request to check, either as a Request object or a dictionary.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizedResponseBrief containing the result of the check.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        response = await self.aauthorize(request, correlation_id=correlation_id)
        if not response.results:
            raise ValueError("No results returned from authorize endpoint")
        result = response.results[0]
        if result.status == "failed":
            raise RuntimeError(f"Authorization failed: {result.error}")
        return result.result or AuthorizedResponseBrief(Decision.DENY)

    async def acheck_detailed(
        self, request: Request | dict[str, Any], correlation_id: str | None = None
    ) -> AuthorizedResponseDetailed:
        """Check the given request with detailed output. Asynchronous version (compatibility wrapper).

        This method provides backward compatibility with the old single-request API.
        It wraps the new batch aauthorize_detailed endpoint.

        Args:
            request: The request to check, either as a Request object or a dictionary.
            correlation_id: Optional correlation ID for tracing the request.
        Returns:
            An AuthorizedResponseDetailed containing the detailed result of the check.
        Raises:
            httpx.HTTPStatusError: If the request fails with a non-2xx status code
        """
        response = await self.aauthorize_detailed(
            request, correlation_id=correlation_id
        )
        if not response.results:
            raise ValueError("No results returned from authorize endpoint")
        result = response.results[0]
        if result.status == "failed":
            raise RuntimeError(f"Authorization failed: {result.error}")
        return result.result or AuthorizedResponseDetailed(Decision.DENY, None, None)

    def close(self):
        """Close the synchronous client connection."""
        with contextlib.suppress(Exception):
            self._sync_client.close()
        TreeTopClient._instance = None  # type: ignore

    async def aclose(self):
        """Close the asynchronous client connection."""
        await self._async_client.aclose()
        self._sync_client.close()
        TreeTopClient._instance = None  # type: ignore


# For typing convenience
RequestLike = Request | dict[str, Any]



from datetime import datetime

import httpx
import pytest
from pytest_httpx import HTTPXMock

from treetop_client.client import TreeTopClient
from treetop_client.models import (Action, Decision, QualifiedId, Request,
                                   Resource, ResourceAttribute, User)


@pytest.fixture(autouse=True)
def client_cleanup():
    # ensure singleton reset between tests
    TreeTopClient().close()
    yield
    TreeTopClient().close()


def make_req(id_suffix: str | None = "1") -> Request:
    return Request(
        principal=User(id=QualifiedId(id="alice"), groups=[]),
        action=Action(id=QualifiedId(id="view")),
        resource=Resource(
            kind="Photo",
            id="42",
            attrs={
                "id": ResourceAttribute.new("42"),
            },
        ),
        id=f"check-{id_suffix}" if id_suffix else None,
    )


# Tests for the new batch authorize endpoint
def test_authorize_single_request_brief(httpx_mock: HTTPXMock):
    """Test authorize endpoint with a single request (brief detail)."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "check-1",
                    "status": "success",
                    "result": {"decision": "Allow"},
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    response = client.authorize(make_req())
    assert len(response) == 1
    assert response.successful == 1
    assert response.failed == 0
    result = response[0]
    assert result.is_success()
    assert result.is_allowed()
    assert result.get_decision() == Decision.ALLOW


def test_authorize_multiple_requests_brief(httpx_mock: HTTPXMock):
    """Test authorize endpoint with multiple requests (brief detail)."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "check-1",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
                {
                    "index": 1,
                    "id": "check-2",
                    "status": "failed",
                    "error": "Evaluation failed: invalid resource",
                },
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 1,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2")]
    response = client.authorize(requests)
    assert len(response) == 2
    assert response.successful == 1
    assert response.failed == 1

    result1 = response[0]
    assert result1.is_success()
    assert result1.is_allowed()

    result2 = response[1]
    assert result2.is_failed()
    assert result2.error == "Evaluation failed: invalid resource"

    # Test get_by_id
    found = response.get_by_id("check-1")
    assert found is not None
    assert found.is_allowed()


def test_authorize_detailed(httpx_mock: HTTPXMock):
    """Test authorize_detailed endpoint with detailed response."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize?detail=full",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "check-1",
                    "status": "success",
                    "result": {
                        "decision": {
                            "Allow": {
                                "policy": {
                                    "literal": 'permit (\n    principal == User::"alice",\n    action in [Action::"view"],\n    resource == Photo::"42"\n);',
                                    "json": {
                                        "action": {
                                            "entities": [
                                                {"id": "view", "type": "Action"}
                                            ],
                                            "op": "in",
                                        },
                                        "conditions": [],
                                        "effect": "permit",
                                        "principal": {
                                            "entity": {"id": "alice", "type": "User"},
                                            "op": "==",
                                        },
                                        "resource": {
                                            "entity": {"id": "42", "type": "Photo"},
                                            "op": "==",
                                        },
                                    },
                                },
                                "version": {
                                    "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                                    "loaded_at": "2025-12-19T15:25:55.384783000Z",
                                },
                            },
                        },
                    },
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    response = client.authorize_detailed(make_req())
    assert len(response) == 1
    result = response[0]
    assert result.is_success()
    assert result.is_allowed()
    assert result.policy_literal() is not None
    policy_lit = result.policy_literal()
    assert policy_lit is not None and 'principal == User::"alice"' in policy_lit
    assert (
        result.version_hash()
        == "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219"
    )
    assert result.version_loaded_at() is not None


def test_authorize_deny(httpx_mock: HTTPXMock):
    """Test authorize endpoint with Deny decision."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "check-1",
                    "status": "success",
                    "result": {"decision": "Deny"},
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    response = client.authorize(make_req())
    assert len(response) == 1
    result = response[0]
    assert result.is_success()
    assert result.is_denied()
    assert result.get_decision() == Decision.DENY


def test_authorize_error(httpx_mock: HTTPXMock):
    """Test authorize endpoint with HTTP error."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={"error": "Invalid request"},
        status_code=400,
    )
    client = TreeTopClient()
    with pytest.raises(httpx.HTTPStatusError):
        client.authorize(make_req())


def test_async_authorize(httpx_mock: HTTPXMock):
    """Test async authorize endpoint."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "check-1",
                    "status": "success",
                    "result": {"decision": "Allow"},
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    import asyncio

    asyncio.set_event_loop(asyncio.new_event_loop())
    response = asyncio.get_event_loop().run_until_complete(
        client.aauthorize(make_req())
    )
    assert len(response) == 1
    result = response[0]
    assert result.is_allowed()


# Backward compatibility tests (old check/check_detailed API)
def test_check_backward_compatibility(httpx_mock: HTTPXMock):
    """Test backward compatibility with old check() method."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": None,
                    "status": "success",
                    "result": {"decision": "Allow"},
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    resp = client.check(make_req(id_suffix=None))
    assert resp.is_allowed()
    assert resp.decision == Decision.ALLOW


def test_check_detailed_backward_compatibility(httpx_mock: HTTPXMock):
    """Test backward compatibility with old check_detailed() method."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize?detail=full",
        json={
            "results": [
                {
                    "index": 0,
                    "id": None,
                    "status": "success",
                    "result": {
                        "decision": {
                            "Allow": {
                                "policy": {
                                    "literal": 'permit (\n    principal == User::"alice",\n    action in [Action::"view"],\n    resource == Photo::"42"\n);',
                                    "json": {
                                        "effect": "permit",
                                        "principal": {
                                            "entity": {"id": "alice", "type": "User"},
                                            "op": "==",
                                        },
                                        "action": {
                                            "entities": [
                                                {"id": "view", "type": "Action"}
                                            ],
                                            "op": "in",
                                        },
                                        "resource": {
                                            "entity": {"id": "42", "type": "Photo"},
                                            "op": "==",
                                        },
                                        "conditions": [],
                                    },
                                },
                                "version": {
                                    "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                                    "loaded_at": "2025-12-16T15:25:55.384783000Z",
                                },
                            },
                        },
                    },
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-16T15:25:55.384783000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    resp = client.check_detailed(make_req(id_suffix=None))
    assert resp.is_allowed()
    assert resp.decision == Decision.ALLOW
    assert resp.policy is not None
    assert "alice" in resp.policy.literal
    assert (
        resp.version_hash()
        == "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219"
    )
    loaded_at = resp.version_loaded_at()
    assert loaded_at is not None
    assert isinstance(loaded_at, datetime)


def test_check_deny_backward_compatibility(httpx_mock: HTTPXMock):
    """Test backward compatibility with old check() method returning Deny."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": None,
                    "status": "success",
                    "result": {"decision": "Deny"},
                }
            ],
            "version": {
                "hash": "c82d116854d77bf689c3d15e167764876dffe869c970bc08ab7c5dacd7726219",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 1,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    resp = client.check(make_req(id_suffix=None))
    assert resp.is_denied()
    assert resp.decision == Decision.DENY


# Batch query tests with index and ID lookups
def test_batch_authorize_lookup_by_index(httpx_mock: HTTPXMock):
    """Test batch authorize with multiple requests - lookup by index."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "req-alice-view",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
                {
                    "index": 1,
                    "id": "req-bob-delete",
                    "status": "success",
                    "result": {"decision": "Deny"},
                },
                {
                    "index": 2,
                    "id": "req-charlie-edit",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 3,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2"), make_req("3")]
    response = client.authorize(requests)

    # Test lookup by index
    assert len(response) == 3
    assert response[0].is_allowed()
    assert response[1].is_denied()
    assert response[2].is_allowed()

    # Verify indexes match
    assert response[0].index == 0
    assert response[1].index == 1
    assert response[2].index == 2


def test_batch_authorize_lookup_by_id(httpx_mock: HTTPXMock):
    """Test batch authorize with multiple requests - lookup by ID."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "req-alice-view",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
                {
                    "index": 1,
                    "id": "req-bob-delete",
                    "status": "success",
                    "result": {"decision": "Deny"},
                },
                {
                    "index": 2,
                    "id": "req-charlie-edit",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 3,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2"), make_req("3")]
    response = client.authorize(requests)

    # Test lookup by ID
    alice_result = response.get_by_id("req-alice-view")
    assert alice_result is not None
    assert alice_result.is_allowed()
    assert alice_result.id == "req-alice-view"

    bob_result = response.get_by_id("req-bob-delete")
    assert bob_result is not None
    assert bob_result.is_denied()
    assert bob_result.id == "req-bob-delete"

    charlie_result = response.get_by_id("req-charlie-edit")
    assert charlie_result is not None
    assert charlie_result.is_allowed()
    assert charlie_result.id == "req-charlie-edit"

    # Test lookup of non-existent ID
    nonexistent = response.get_by_id("req-nonexistent")
    assert nonexistent is None


def test_batch_authorize_detailed_lookup_by_index(httpx_mock: HTTPXMock):
    """Test batch authorize_detailed with multiple requests - lookup by index."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize?detail=full",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "req-1",
                    "status": "success",
                    "result": {
                        "decision": {
                            "Allow": {
                                "policy": {
                                    "literal": "permit (...);",
                                    "json": {
                                        "effect": "permit",
                                    },
                                },
                                "version": {
                                    "hash": "hash1",
                                    "loaded_at": "2025-12-19T00:14:38.577289000Z",
                                },
                            }
                        }
                    },
                },
                {
                    "index": 1,
                    "id": "req-2",
                    "status": "success",
                    "result": {
                        "decision": {
                            "Deny": {
                                "version": {
                                    "hash": "hash2",
                                    "loaded_at": "2025-12-19T00:14:38.577289000Z",
                                }
                            }
                        }
                    },
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 2,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2")]
    response = client.authorize_detailed(requests)

    # Test lookup by index
    assert len(response) == 2
    assert response[0].is_allowed()
    assert response[0].policy_literal() is not None
    assert response[0].version_hash() == "hash1"

    assert response[1].is_denied()
    assert response[1].policy_literal() is None
    assert response[1].version_hash() == "hash2"


def test_batch_authorize_detailed_lookup_by_id(httpx_mock: HTTPXMock):
    """Test batch authorize_detailed with multiple requests - lookup by ID."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize?detail=full",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "photo-allow",
                    "status": "success",
                    "result": {
                        "decision": {
                            "Allow": {
                                "policy": {
                                    "literal": "permit (...);",
                                    "json": {
                                        "effect": "permit",
                                    },
                                },
                                "version": {
                                    "hash": "hash1",
                                    "loaded_at": "2025-12-19T00:14:38.577289000Z",
                                },
                            }
                        }
                    },
                },
                {
                    "index": 1,
                    "id": "video-deny",
                    "status": "success",
                    "result": {
                        "decision": {
                            "Deny": {
                                "version": {
                                    "hash": "hash2",
                                    "loaded_at": "2025-12-19T00:14:38.577289000Z",
                                }
                            }
                        }
                    },
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 2,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [
        Request(
            principal=User(id=QualifiedId(id="alice"), groups=[]),
            action=Action(id=QualifiedId(id="view")),
            resource=Resource(
                kind="Photo",
                id="42",
                attrs={"id": ResourceAttribute.new("42")},
            ),
            id="photo-allow",
        ),
        Request(
            principal=User(id=QualifiedId(id="bob"), groups=[]),
            action=Action(id=QualifiedId(id="view")),
            resource=Resource(
                kind="Video",
                id="99",
                attrs={"id": ResourceAttribute.new("99")},
            ),
            id="video-deny",
        ),
    ]
    response = client.authorize_detailed(requests)

    # Test lookup by ID
    photo_result = response.get_by_id("photo-allow")
    assert photo_result is not None
    assert photo_result.is_allowed()
    assert photo_result.policy_literal() is not None
    assert photo_result.version_hash() == "hash1"

    video_result = response.get_by_id("video-deny")
    assert video_result is not None
    assert video_result.is_denied()
    assert video_result.policy_literal() is None
    assert video_result.version_hash() == "hash2"


def test_batch_authorize_mixed_success_and_failure(httpx_mock: HTTPXMock):
    """Test batch authorize with both successful and failed results."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "req-ok",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
                {
                    "index": 1,
                    "id": "req-error",
                    "status": "failed",
                    "error": "Invalid resource kind",
                },
                {
                    "index": 2,
                    "id": "req-ok2",
                    "status": "success",
                    "result": {"decision": "Deny"},
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 2,
            "failed": 1,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2"), make_req("3")]
    response = client.authorize(requests)

    # Verify counts
    assert response.successful == 2
    assert response.failed == 1
    assert len(response) == 3

    # Test lookup by index with error
    assert response[0].is_success()
    assert response[0].is_allowed()

    assert response[1].is_failed()
    assert response[1].error == "Invalid resource kind"
    assert response[1].result is None

    assert response[2].is_success()
    assert response[2].is_denied()

    # Test lookup by ID
    found_ok = response.get_by_id("req-ok")
    assert found_ok is not None
    assert found_ok.is_allowed()

    found_error = response.get_by_id("req-error")
    assert found_error is not None
    assert found_error.is_failed()

    found_ok2 = response.get_by_id("req-ok2")
    assert found_ok2 is not None
    assert found_ok2.is_denied()


def test_batch_authorize_iteration(httpx_mock: HTTPXMock):
    """Test iterating over batch authorize results."""
    httpx_mock.add_response(
        method="POST",
        url="http://localhost:9999/api/v1/authorize",
        json={
            "results": [
                {
                    "index": 0,
                    "id": "req-1",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
                {
                    "index": 1,
                    "id": "req-2",
                    "status": "success",
                    "result": {"decision": "Deny"},
                },
                {
                    "index": 2,
                    "id": "req-3",
                    "status": "success",
                    "result": {"decision": "Allow"},
                },
            ],
            "version": {
                "hash": "abc123",
                "loaded_at": "2025-12-19T00:14:38.577289000Z",
            },
            "successful": 3,
            "failed": 0,
        },
        status_code=200,
    )
    client = TreeTopClient()
    requests = [make_req("1"), make_req("2"), make_req("3")]
    response = client.authorize(requests)

    # Test iteration
    decisions = [result.get_decision() for result in response]
    assert decisions == [Decision.ALLOW, Decision.DENY, Decision.ALLOW]

    # Test iteration with filter
    allowed_count = sum(1 for result in response if result.is_allowed())
    assert allowed_count == 2



import pytest

from treetop_client.models import (
    Action,
    Group,
    QualifiedId,
    Request,
    Resource,
    ResourceAttribute,
    User,
)


def test_qualified_id_and_group():
    q = QualifiedId(id="alice", namespace=["App"])
    g = Group(id=q)
    u = User(id=q, groups=[g])
    req = Request(
        principal=u,
        action=Action(id=QualifiedId(id="x")),
        resource=Resource(
            kind="Photo",
            id="1",
            attrs={"id": ResourceAttribute.new("1")},
        ),
    )
    api = req.to_api()
    assert api["principal"]["User"]["id"] == "alice"
    assert api["principal"]["User"]["namespace"] == ["App"]
    assert api["principal"]["User"] == {
        "id": "alice",
        "namespace": ["App"],
        "groups": [{"id": "alice", "namespace": ["App"]}],
    }


def test_resource_empty_attrs():
    with pytest.raises(ValueError):
        Resource(kind="Photo", id="1", attrs={})


def test_user_no_colon():
    with pytest.raises(ValueError):
        User(id=QualifiedId(id="bad:user"))


def test_group_no_colon():
    with pytest.raises(ValueError):
        Group(id=QualifiedId(id="bad:group", namespace=["App"]))


def test_action_no_colon():
    with pytest.raises(ValueError):
        Action(id=QualifiedId(id="bad:action"))


def test_user_with_namespace():
    q = QualifiedId(id="alice", namespace=["App"])
    u = User(id=q, groups=[])
    assert u.to_api() == {
        "id": "alice",
        "namespace": ["App"],
        "groups": [],
    }


def test_group_with_namespace():
    g = Group(id=QualifiedId(id="group1", namespace=["App"]))
    assert g.to_api() == {"id": "group1", "namespace": ["App"]}


def test_action_with_namespace():
    q = QualifiedId(id="edit", namespace=["App"])
    a = Action(id=q)
    assert a.to_api() == {"id": "edit", "namespace": ["App"]}


def test_user_new():
    u = User.new(id="alice", namespace=["App"], groups=["group1", "group2"])
    assert u.to_api() == {
        "id": "alice",
        "namespace": ["App"],
        "groups": [
            {"id": "group1", "namespace": ["App"]},
            {"id": "group2", "namespace": ["App"]},
        ],
    }


def test_group_new():
    g = Group.new(id="group1", namespace=["App"])
    assert g.to_api() == {"id": "group1", "namespace": ["App"]}


def test_action_new():
    a = Action.new(id="edit", namespace=["App"])
    assert a.to_api() == {"id": "edit", "namespace": ["App"]}
